/**
 * @file c2n.h
 * Interface definitions related to the Commodore C2N storage format
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2001,2004 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

/** Pulse widths of symmetric waves */
enum pulse
{
  Pause = 'A',		/**< break in the pulse stream */
  Short = 'B',		/**< short pulse */
  Medium = 'C',		/**< medium pulse */
  Long = 'D'		/**< long pulse */
};

/** Asymmetric waves (negative pulse width first, then positive) */
enum pulse_asymm
{
  PauseShort = 'E',	/**< pause followed by short pulse */
  ShortShort = 'F',	/**< short and short pulse */
  MediumShort = 'G',	/**< medium and short pulse */
  LongShort = 'H',	/**< long and short pulse */
  PauseMedium = 'I',	/**< pause followed by medium pulse */
  ShortMedium = 'J',	/**< short and medium pulse */
  MediumMedium = 'K',	/**< medium and medium pulse */
  LongMedium = 'L',	/**< long and medium pulse */
  PauseLong = 'M',	/**< pause followed by long pulse */
  ShortLong = 'N',	/**< short and long pulse */
  MediumLong = 'O',	/**< medium and long pulse */
  LongLong = 'P'	/**< long and long pulse */
};
/** Errors related to the pulse stream */
enum pulse_error
{
  OutOfSync = 1,/**< out of synchronisation */
  Parity,	/**< error in the parity bit */
  Checksum,	/**< checksum mismatch */
  NoFirst,	/**< first copy of a block is missing */
  NoSecond,	/**< second copy of a block is missing */
  LongBlock,	/**< block too long */
  ShortBlock,	/**< block too short */
  Unexpected,	/**< unexpected header tag */
  Mismatch,	/**< mismatch between first and second copy */
  Countdown	/**< unrecognized header countdown sequence */
};
/** Commodore C2N tape header identifier tags */
enum c2n_tag
{
  tBasic = 1,      /**< relocatable (BASIC) program */
  tDataBlock = 2,  /**< actual data block of a data file */
  tML = 3,         /**< absolute program */
  tDataHeader = 4, /**< data file header */
  tEnd = 5         /**< end-of-tape header */
};

/** Pulse stream reader
 * @return	a pulse from the stream, or Pause at end of stream
 */
typedef enum pulse (*pulse_r_t) (void);
/** Pulse stream writer
 * @param p	the pulse to be written
 */
typedef void (*pulse_w_t) (enum pulse p);

/** Error reporter
 * @param pec	the unexpected pulse or error code, or character (| 256)
 * @param block	the block in which the error occurred (0=first header)
 * @param pos	number of bytes decoded from the stream
 * @return	nonzero on fatal error
 */
typedef unsigned (*pulse_error_t) (unsigned pec, unsigned block, unsigned pos);

/** Flag: verbose mode */
extern unsigned verbose;
