/**
 * @file pcm_in.c
 * Pulse Code Modulated audio input
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2004 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#if defined USE_PCM && defined USE_TAP
# include <stdio.h>
# include "pcm_in.h"
# include <string.h>

/** PCM audio decoder context */
struct pcm_in
{
  /** Input stream */
  FILE* f;
  /** Input stream position (for diagnostics) */
  long fpos;
  /** Number of audio channels (0=no "fmt " chunk seen) */
  unsigned channels;
  /** Number of sample bytes per channel (0=read no more data) */
  unsigned bytes;
  /** Remaining length of the RIFF block in bytes */
  unsigned long rifflen;
  /** Remaining length of the data block in bytes */
  unsigned long datalen;
  /** Threshold for detecting a low-to-high transition (1..255) */
  int thresh1;
  /** Threshold for detecting a high-to-low transition (0..254) */
  int thresh0;
  /** Flag: inverted phase */
  int inv;
  /** Last audio sample processed */
  int last;
  /** Length of the current cycle */
  double cyclen;
  /** Length of a sample at 125 kHz */
  double unit;
  /** Last transition detected */
  enum { NEGATIVE, POSITIVE } level;
};

/** The PCM audio decoder context */
static struct pcm_in in;
/** Test: is the PCM input stream open? */
#define pcm_open (!!in.bytes)

/** Read a character from the RIFF block
 * @return	the character, or EOF
 */
static int
getriffc (void)
{
  int c;
  if (!in.rifflen || !pcm_open)
    return EOF;
  in.rifflen--;
  c = getc (in.f);
  if (c != EOF)
    in.fpos++;
  return c;
}

/** Read a header from a RIFF file
 * @param hdr	(output) the header tag (4 bytes)
 * @param len	(output) length of the header
 * @return	zero on success, nonzero on failure
 */
static int
read_header (unsigned char* hdr,
	     unsigned long* hdrlen)
{
  unsigned len;
  int i, c;
  for (i = 0; i < 4; i++) {
    if ((c = getriffc ()) == EOF)
      return EOF;
    hdr[i] = c;
  }
  len = 0;
  for (i = 0; i < 32; i += 8) {
    if ((c = getriffc ()) == EOF)
      return EOF;
    len |= c << i;
  }
  *hdrlen = len;
  return 0;
}

/** Skip the specified number of following bytes of a file.
 * @param f	file
 * @param len	number of bytes to be skipped
 * @return	zero on success, nonzero on failure
 */
static int
read_ignore (unsigned long len)
{
  if (len > in.rifflen)
    len = in.rifflen;
  if (!fseek (in.f, len, SEEK_CUR)) {
    in.rifflen -= len;
    in.fpos += len;
  }
  else
    for (; len; len--)
      if (getriffc () == EOF)
	return EOF;
  return 0;
}

/** Initialize the PCM audio decoder
 * @param f		the input stream
 * @param bias		DC level of the input (normally 128)
 * @param threshold	hysteresis threshold for detecting transitions;
 *			if <0, invert the phase
 */
void
pcm_in_init (FILE* f,
	     int bias,
	     int threshold)
{
  in.f = f ? f : stdin;
  in.fpos = 0L;
  in.bytes = 1; /* pcm_open = TRUE */
  in.inv = threshold < 0;
  if (threshold < 0)
    threshold = -threshold;
  in.thresh1 = bias + threshold;
  in.thresh0 = bias - threshold;
}

/** Clean up the PCM audio decoder */
void
pcm_in_cleanup (void)
{
  if (in.f && in.f != stdin)
    fclose (in.f);
  in.f = 0;
  in.fpos = 0L;
  in.channels = in.bytes = 0;
  in.datalen = in.rifflen = 0UL;
  in.last = 128;
  in.cyclen = in.unit = 0;
  in.level = NEGATIVE;
}

/** Find next PCM audio chunk
 * @return	zero on success, nonzero on failure
 */
int
pcm_in_next (void)
{
  /** the "RIFF" tag */
  static const unsigned char tag_riff[4] = { 0x52, 0x49, 0x46, 0x46 };
  /** the "WAVE" tag */
  static const unsigned char tag_wave[4] = { 0x57, 0x41, 0x56, 0x45 };
  /** the "fmt " tag */
  static const unsigned char tag_fmt[4] = { 0x66, 0x6d, 0x74, 0x20 };
  /** the "data" tag */
  static const unsigned char tag_data[4] = { 0x64, 0x61, 0x74, 0x61 };
  int i;
  unsigned char hdr[4];
  if (!pcm_open)
    return 3; /* file not open */
 again:
  in.channels = in.bytes = 0;
  in.datalen = 0UL;
  in.bytes = 1; /* pcm_open = TRUE */ 
  in.cyclen = in.unit = 0;
  in.level = NEGATIVE;
  if (!in.rifflen) {
    in.rifflen = 8UL;
    if (read_header (hdr, &in.rifflen)) {
    eof:
      in.bytes = 0; /* pcm_open = FALSE */
      return 2; /* unexpected EOF */
    }
    if (memcmp (hdr, tag_riff, 4)) {
      in.bytes = 0; /* pcm_open = FALSE */
      return 1; /* no RIFF header */
    }
    if (in.rifflen < 4) {
    ignore_riff:
      if (read_ignore (in.rifflen))
	goto eof;
      in.rifflen = 0UL;
      goto again;
    }
    if (4 != fread (hdr, 1, 4, in.f))
      return 2; /* unexpected EOF */
    in.rifflen -= 4, in.fpos += 4;
    if (memcmp (hdr, tag_wave, 4))
      goto ignore_riff; /* no "WAVE" identifier: skip to next RIFF header */
  }

  /* search for "fmt " chunk */
  while (in.rifflen > 8) {
    unsigned long len;
    if (read_header (hdr, &len))
      return 2; /* unexpected EOF */
    if (!memcmp (hdr, tag_fmt, 4)) {
      unsigned long l;
      /* found "fmt " chunk */
      if (len < 16UL) {
      ignore_format:
	fputs ("ignoring non-PCM RIFF WAVE format\n", stderr);
	goto ignore_riff; /* chunk too small */
      }
      len -= 16UL;
      /* Format category */
      if (getriffc () != 1 ||
	  getriffc () != 0)
	goto ignore_format; /* not PCM format */
      /* Number of channels */
      in.channels = getriffc ();
      in.channels |= getriffc () << 8;
      if (!in.channels) {
	fputs ("ignoring chunk with no audio channels\n", stderr);
	goto ignore_riff;
      }
      /* Sampling rate */
      l = getriffc ();
      l |= getriffc () << 8;
      l |= getriffc () << 16;
      l |= getriffc () << 24;
      if (!l) {
	fputs ("sampling rate must not be zero\n", stderr);
	goto ignore_riff;
      }
      in.unit = (double) 125000 / l;
      /* Average bytes per second (ignored) */
      getriffc (); getriffc (); getriffc (); getriffc ();
      /* Block alignment */
      l = getriffc ();
      l = getriffc () << 8;
      /* Bits per channel */
      i = getriffc ();
      i |= getriffc () << 8;
      if (i < 1 || i > 64) {
	fputs ("unsupported number of bits per channel\n", stderr);
	goto ignore_riff;
      }
      i = (i + 7) / 8; /* convert to bytes per channel */
      if (l % (i * in.channels)) {
	fputs ("inconsistent block alignment\n", stderr);
	goto ignore_riff; /* inconsistent block alignment size */
      }
      in.bytes = i;
    }
    if (!memcmp (hdr, tag_data, 4)) {
      /* found "data" chunk */
      /* to do: if no "fmt " chunk, goto again */
      /* check: len == rifflen || len == 0xffffffff */
      if (!in.channels) {
	fputs ("ignoring data chunk without fmt\n", stderr);
	goto ignore_riff;
      }
      in.datalen = len;
      return 0;
    }
    if (read_ignore (len))
      return 2; /* unexpected EOF */
  }
  goto ignore_riff;
}

/* Read an unsigned 8-bit sample from the first channel
 * @return	the converted sample, or EOF
 */
static int
read_sample (void)
{
  int c, sample = 0;
  unsigned long l = in.bytes * in.channels;

  if (in.datalen < l || in.rifflen < l)
    return EOF;
  in.datalen -= l, in.rifflen -= l;

  /* For skipping a small number of bytes, getc() appears much faster
     than fseek().  Thus this function does not call read_ignore(). */

  l -= in.bytes;

  /* Read the most significant byte of the first channel */
  for (c = in.bytes; c--; )
    if ((sample = getc (in.f)) == EOF)
      return EOF;

  /* Skip the remaining channels */
  while (l--)
    if (getc (in.f) == EOF)
      return EOF;

  in.fpos += in.bytes * in.channels;

  /* Convert the byte into unsigned format if needed */
  if (in.bytes > 1)
    sample ^= 128;

  /* Invert the phase if needed */
  if (in.inv)
    sample ^= 255;

  return sample;
}

/** Extract a cycle from a PCM audio file
 * @return	the cycle width (unit: 8 s), 255 (pause) or EOF
 */
int
pcm_in_read (void)
{
  int c;
  if (!pcm_open)
    return EOF;
  /* The algorithm has been adapted from TAPir 0.05b by Andreas Andersson. */
  for (;; in.cyclen += in.unit, in.last = c) {
    /* Read samples until the next positive-to-negative transition.
     * Return the time elapsed from the last transition.
     */
    c = read_sample ();
    if (c == EOF)
      return EOF;
    if (in.level == NEGATIVE) {
      /* Negative-to-positive transitions are not reported. */
      if (c >= in.thresh1)
	in.level = POSITIVE;
      continue;
    }
    /* Watch for positive-to-negative transitions */
    if (c <= in.thresh0) {
      /* Determine the intersection point by linear interpolation */
      double delta = in.unit * (in.last - in.thresh0) / (in.last - c);
      in.level = NEGATIVE;
      in.last = c;
      in.cyclen += delta;
      c = in.cyclen > 255
	? 255 /* pause */
	: in.cyclen;
      /* This many samples have been seen of the next cycle */
      in.cyclen = in.unit - delta;
      return c;
    }
  }
}

/** Get the PCM file position
 * @return	byte offset from the beginning of the file
 */
long
pcm_in_pos (void)
{
  return in.fpos;
}

#endif /* USE_PCM && USE_TAP */
