/**
 * @file serialw.c
 * The Win32 serial line handler
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2001,2002,2004,2006 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include "c2n.h"
#include "serial.h"
#ifdef HAS_SERIAL
/** The name of the serial device */
static const char* serialdev;
# include <stdio.h>
# include <string.h>
# include <errno.h>
/** file handle for the serial device (INVALID_HANDLE_VALUE=none) */
HANDLE serialfd = INVALID_HANDLE_VALUE;
/** previous terminal input/output settings for the serial line */
static DCB tio;
/** previous timeout settings for the serial line */
static COMMTIMEOUTS tio2;
/** new terminal input/output settings for the serial line */
static DCB newtio;
/** new timeout settings for the serial line */
static COMMTIMEOUTS newtio2;
/** Report a Windows error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  char* msg;
  long err = GetLastError ();
  FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		 FORMAT_MESSAGE_FROM_SYSTEM | 
		 FORMAT_MESSAGE_IGNORE_INSERTS,
		 0,
		 err,
		 0, /* default language */
		 (LPTSTR) &msg,
		 0,
		 0);
  AnsiToOem (msg, msg);
  fputs ("-c ", stderr);
  fputs (serialdev, stderr);
  fputs (": ", stderr);
  fputs (op, stderr);
  fprintf (stderr, " failed with code %ld: ", err);
  fputs (msg, stderr);
  LocalFree (msg);
}

/** input/output buffer */
static char buf[8192];
/** number of bytes read from or written to buf */
static unsigned datain;
/** number of bytes in buf */
static unsigned datalen;
/** nonzero=fast mode (unsafe with USB RS-232 adapters) */
static int serial_fast;

/** Set up the serial communications
 * @param dev		serial device driver name
 * @param fast		nonzero=fast mode (unsafe with USB RS-232 adapters)
 * @return		zero on success, nonzero on failure
 */
int
serial_init (const char* dev, int fast)
{
  serial_fast = fast;

  if ((serialfd = CreateFile (serialdev = dev, GENERIC_WRITE | GENERIC_READ,
			      0, 0, OPEN_EXISTING, 0, 0)) ==
      INVALID_HANDLE_VALUE) {
    report_error ("CreateFile");
    return 1;
  }

  if (!GetCommState (serialfd, &tio)) {
    report_error ("GetCommState");
  c2n232fail:
    CloseHandle (serialfd);
    serialfd = INVALID_HANDLE_VALUE;
    return 2;
  }
  else if (!GetCommTimeouts (serialfd, &tio2)) {
    report_error ("GetCommTimeouts");
    goto c2n232fail;
  }
  else {
    memcpy (&newtio, &tio, sizeof newtio);
    memcpy (&newtio2, &tio2, sizeof newtio2);
    newtio.BaudRate = CBR_38400;
    newtio.fBinary = 1;
    newtio.fParity = newtio.fOutxCtsFlow = newtio.fOutxDsrFlow = 0;
    newtio.fDtrControl = DTR_CONTROL_DISABLE;
    newtio.fDsrSensitivity = 0;
    newtio.fOutX = 1, newtio.fInX = 0;
    newtio.fOutxCtsFlow = !fast; /* CTS/RTS output handshaking */
    newtio.fErrorChar = newtio.fNull = 0;
    newtio.fRtsControl = RTS_CONTROL_DISABLE;
    newtio.fAbortOnError = 0;
    newtio.ByteSize = 8;
    newtio.Parity = NOPARITY;
    newtio.StopBits = ONESTOPBIT;
    newtio.XonChar = 021; /* DC1, ctrl-q */
    newtio.XoffChar = 023; /* DC3, ctrl-s */

    newtio2.ReadIntervalTimeout = 100; /* 1/10th second */
    newtio2.ReadTotalTimeoutConstant = 100; /* 1/10th second */
    newtio2.ReadTotalTimeoutMultiplier =
      newtio2.WriteTotalTimeoutMultiplier =
      newtio2.WriteTotalTimeoutConstant = 0;

    if (!SetCommState (serialfd, &newtio)) {
      report_error ("SetCommState");
      goto c2n232fail;
    }
    else if (!SetCommTimeouts (serialfd, &newtio2)) {
      report_error ("SetCommTimeouts");
      goto c2n232fail;
    }
    else {
      unsigned long len;
      /* send a NUL character and await a response */
      if (!WriteFile (serialfd, "", 1, &len, 0)) {
	report_error ("WriteFile");
	goto c2n232fail;
      }
      if (!ReadFile (serialfd, buf, sizeof buf, &len, 0)) {
	report_error ("ReadFile");
	goto c2n232fail;
      }
      if (!len || buf[len - 1] != '0') {
	fputs ("-c ", stderr);
	fputs (serialdev, stderr);
	fputs (len
	       ? ": unexpected response from C2N232 device\n"
	       : ": no response from C2N232 device\n",
	       stderr);
	goto c2n232fail;
      }

      newtio2.ReadIntervalTimeout = 100; /* 1/10th second */
      newtio2.ReadTotalTimeoutConstant = 2000; /* 2 seconds */
      if (!SetCommTimeouts (serialfd, &newtio2)) {
	report_error ("SetCommTimeouts");
	goto c2n232fail;
      }
    }
  }
  return 0;
}

void
serial_cleanup (void)
{
  if (serialfd != INVALID_HANDLE_VALUE) {
    PurgeComm (serialfd, PURGE_TXCLEAR | PURGE_RXCLEAR);
    SetCommState (serialfd, &tio);
    SetCommTimeouts (serialfd, &tio2);
    SetCommBreak (serialfd);
    Sleep (20); /* send break signal for at least 20 milliseconds */
    ClearCommBreak (serialfd);
    CloseHandle (serialfd);
    serialfd = INVALID_HANDLE_VALUE;
  }
}

/** Read a pulse from the serial line
 * @return	the read pulse
 */
enum pulse
sp_read (void)
{
  char c;
  if (!serial_open)
    return Pause;
 again:
  if (datain < datalen)
    c = buf[datain++];
  else {
    unsigned long len;
    if (!ReadFile (serialfd, buf, sizeof buf, &len, 0) || !len) {
      serial_cleanup ();
      return Pause;
    }
    datalen = len;
    datain = 0;
    goto again;
  }
  switch (c) {
  case Pause: case Short: case Medium: case Long:
    return c;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n",
	   (unsigned char) c);
  goto again;
}

/** Empty the serial buffers */
void
serial_empty (void)
{
  datain = 0, datalen = 0;
}

/** Flush the serial output buffer */
void
serial_flush (void)
{
}

/** Send a command to the C2N232 device
 * @param buf	the command block
 * @param len	the length of the block
 * @return	zero on success, nonzero on failure
 */
int
serial_command (unsigned char* buf,
		unsigned len)
{
  unsigned long length;
  if (!WriteFile (serialfd, buf, len, &length, 0)) {
    report_error ("WriteFile");
    return 1;
  }
  if (!ReadFile (serialfd, buf + 1, 1, &length, 0)) {
    report_error ("ReadFile");
    return 2;
  }

  if ((buf[1] & ~0x30) == buf[0])
    return 0;
  fprintf (stderr, "C2N232 responded 0x%02x; expected 0x%02x\n",
	   buf[1], buf[0] | 0x30);
  return 3;
}

/** Disable the serial read timeout
 * @return	zero on success, nonzero on failure
 */
int
serial_timeout (void)
{
  newtio2.ReadIntervalTimeout = 0; /* no waiting at reading */
  if (!SetCommTimeouts (serialfd, &newtio2)) {
    report_error ("SetCommTimeouts");
    return 1;
  }
  return 0;
}

/** Make the C2N232 go idle */
void
serial_go_idle (void)
{
  unsigned long len;
  if (!FlushFileBuffers (serialfd))
    report_error ("FlushFileBuffers");
  if (!serial_fast)
    Sleep (250);
  WriteFile (serialfd, "", 1, &len, 0);
}

/** Write a pulse to the serial line or to standard output
 * @param p	the pulse to be written
 */
void
p_write (enum pulse p)
{
  if (serial_open) {
    buf[datain++] = p;
    if (p == Pause || datain == sizeof buf) {
      unsigned long len;
      WriteFile (serialfd, buf, datain, &len, 0);
      datain = 0;
    }
  }
  else
    putchar (p);
}
#endif /* HAS_SERIAL */
